//-------------------------------------------------------------------------
// play-v6 - music player for ATmega328P @ 16 MHz
// top-level functions
// created 9 Dec 2013 by Enrico Colombini <erix@erix.it>
// license: CC Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)
// http://creativecommons.org/licenses/by-sa/4.0/
//-------------------------------------------------------------------------

#include "play_v6.h"
#include "play.h"
#include "pwm.h"
#include "userial.h"
#include "music.h"
#include "interface.h"
#include <stddef.h>
#include <stdint.h>
#include <stdbool.h>
#include <ctype.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/crc16.h>

//-------------------------------------------------------------------------
// local constants

//load command results and their descriptions
enum LoadResult {
    LOAD_OK = 0, LOAD_ERR_SIZE, LOAD_ERR_CRC, LOAD_ERR_TIMEOUT
};

const char loadResultErrSize[] PROGMEM = "load: music size too large\n";
const char loadResultErrCrc[] PROGMEM = "load: CRC error\n";
const char loadResultErrTimeout[] PROGMEM = "load: timeout\n";

//-------------------------------------------------------------------------
// local variables

//music is played from here
static uint8_t musicBuf[MAX_MUSIC_LENGTH];

//index of last built-in music copied to musicBuf
static uint8_t lastBuiltinMusicIdx = 0;

//-------------------------------------------------------------------------
// local functions

static void    CommandLoop(void);
static bool    CheckButtons(void);
static uint8_t SelectBuiltinMusic(int8_t idx);
static bool    PlayCurrentMusic(void);
static void    AutoPlay(bool advance);
static bool    InterPlayWait(uint32_t pwmCycles);

static void    LoadMusicAndReply(void);
static enum LoadResult LoadMusic(void);

static uint8_t LoadWaitSerialByteIfOk(uint8_t *crc8, enum LoadResult *result);
static PGM_P   GetLoadReplyString(enum LoadResult r);

static uint8_t GetSerialCommand(void);
static bool    SerialWaitByteWithTimeout(uint8_t *receivedByte, uint16_t timeoutCycles);
static void    SerialSendPgmString(PGM_P ps);

//-------------------------------------------------------------------------
// actual main program
// (called from external main() for portability to the idiosincratic Arduino IDE) +++

void PlayerMain(void)
{
    //no interrupts, please, it's all synchronous
    cli();

    //init serial port
    UART0_Init(UART0_BAUD_FACTOR(CPU_CLOCK, SERIAL_BAUD_RATE));

    //setup button port bits and debounce, read current state
    InitInterface();

    //start synchronized PWMs, ramp-up slowly to silence level (analog zero)
    InitAudioPWM();
    RampUpToSilence();

    //copy first built-in music to RAM buffer, remember its index
    lastBuiltinMusicIdx = SelectBuiltinMusic(0);

#if TEST_MODE_NOSYNC

    //test mode: play directly
    Play(musicBuf);

#else //normal mode

    //ignore any serial data sent during initialization
    UART0_FlushRx();

    //if Play button is pressed, start in autoplay mode
    CopyButtonHardwareState();
    if (IsButtonPressed(PLAY_BUTTON_MASK))
    {
        AutoPlay(true);
    }

    //start waiting for commands (this call never returns)
    CommandLoop();

#endif

}

//-------------------------------------------------------------------------
// wait for commands, execute them

static void CommandLoop(void)
{
    while (true)
    {
        //ensure correct timing and button debounce
        SyncToPWM();

        //presume no serial command
        uint8_t serialCmd = SERIAL_CMD_NONE;

        if (CheckButtons())
        {
            //Play button pressed: play music once
            PlayCurrentMusic();
        }
        else
        {
            //Play button not pressed, check for serial command
            serialCmd = GetSerialCommand();
        }

        //execute serial command, if any
        //(will do nothing if a button press has been served)
        switch (serialCmd)
        {
            case SERIAL_CMD_QUERY:
                //status check: ready to accept commands
                UART0_SendByte(SERIAL_REPLY_OK);
                break;

            case SERIAL_CMD_LOAD:
                //load music from serial, will send feedback
                LoadMusicAndReply();
                break;

            case SERIAL_CMD_PLAY_TOGGLE:
                //play music once, handle next/prev commands
                UART0_SendByte(SERIAL_REPLY_OK);
                PlayCurrentMusic();
                break;

            case SERIAL_CMD_REPEAT:
                //play same music repeatedly
                UART0_SendByte(SERIAL_REPLY_OK);
                AutoPlay(false);
                break;

            case SERIAL_CMD_AUTOPLAY:
                //play all built-in music repeatedly
                UART0_SendByte(SERIAL_REPLY_OK);
                AutoPlay(true);
                break;

            case SERIAL_CMD_NEXT:
                //select next music (index will wrap if out of range)
                UART0_SendByte(SERIAL_REPLY_OK);
                lastBuiltinMusicIdx = SelectBuiltinMusic(lastBuiltinMusicIdx + 1);
                break;

            case SERIAL_CMD_PREV:
                //select previous music (index will wrap if out of range)
                UART0_SendByte(SERIAL_REPLY_OK);
                lastBuiltinMusicIdx = SelectBuiltinMusic(lastBuiltinMusicIdx - 1);
                break;

            default:
                //SERIAL_CMD_NONE or invalid command: just ignore
                break;

        } //(switch(serialCmd))

    } //(loop forever)
}

//-------------------------------------------------------------------------
// debounce and read buttons,
// change to next/previous music if requested
// return true if Play button has been pressed

static bool CheckButtons(void)
{
    bool playPressed = false;

    //read and debounce buttons
    uint8_t changedButtons = UpdateButtons();

    //check for button requests
    if (HasButtonBeenPressed(changedButtons, PLAY_BUTTON_MASK))
    {
        playPressed = true;
    }
    else if (HasButtonBeenPressed(changedButtons, NEXT_BUTTON_MASK))
    {
        //select next music (index will wrap if out of range)
        lastBuiltinMusicIdx = SelectBuiltinMusic(lastBuiltinMusicIdx + 1);
    }
    else if (HasButtonBeenPressed(changedButtons, PREV_BUTTON_MASK))
    {
        //select previous music (index will wrap if out of range)
        lastBuiltinMusicIdx = SelectBuiltinMusic(lastBuiltinMusicIdx - 1);
    }

    return playPressed;
}

//-------------------------------------------------------------------------
// select built-in music: wrap idx to range, copy music to the RAM buffer
// return idx of selected music

static uint8_t SelectBuiltinMusic(int8_t idx)
{
    //be sure music index is valid
    int8_t musicCount = (int8_t)pgm_read_byte_near(&builtinMusicCount);
    if (idx >= musicCount)
    {
        idx = 0;
    }
    else if (idx < 0)
    {
        idx = musicCount - 1;
    }

    //get address of requested music
    uint16_t addr = pgm_read_word_near(&builtinMusicList[idx]);
    //first 2 bytes are the music length
    uint16_t len = pgm_read_word_near(addr);
    //copy music to RAM buffer as current music
    memcpy_P(musicBuf, (const uint8_t *)(addr + 2), len);

    //index of current music in RAM buffer
    return idx;
}

//-------------------------------------------------------------------------
// play current music (and others if requested):
// - if stopped with previous/next request, change music and play again
// - if music finishes playing, return true
// - if stopped with button or serial, return false

static bool PlayCurrentMusic(void)
{
    enum PlayState exitState;

    do
    {
        exitState = Play(musicBuf);

        if (exitState == PLAY_EXIT_NEXT)
        {
            //play next music (index will wrap if out of range)
            lastBuiltinMusicIdx = SelectBuiltinMusic(lastBuiltinMusicIdx + 1);
        }
        else if (exitState == PLAY_EXIT_PREV)
        {
            //play previous music (index will wrap if out of range)
            lastBuiltinMusicIdx = SelectBuiltinMusic(lastBuiltinMusicIdx - 1);
        }

    } while((exitState != PLAY_EXIT_DONE) && (exitState != PLAY_EXIT_STOP));

    //set all buttons as pressed to avoid spurious pressing (must release first)
    SetButtonState(PLAY_BUTTON_MASK | NEXT_BUTTON_MASK | PREV_BUTTON_MASK);

    //return true if not stopped
    return (exitState == PLAY_EXIT_DONE);
}

//-------------------------------------------------------------------------
// play repeatedly, with REPEAT_PAUSE_MS between pieces
// exit when stopped with button or serial
// advance to next song (wrapping) if required

static void AutoPlay(bool advance)
{
    //play, exit loop if music stopped
    while (PlayCurrentMusic())
    {
        //music finished naturally, pause for given time
        bool waitOk = InterPlayWait(MS_TO_PWM_CYCLES(REPEAT_PAUSE_MS));

        //move to next music if requested and not stopped
        if (advance && waitOk)
        {
            lastBuiltinMusicIdx = SelectBuiltinMusic(lastBuiltinMusicIdx + 1);
        }
    }
}

//-------------------------------------------------------------------------
// wait for the given number of PWM cycles,
// check buttons and serial,
// exit on play/stop button or SERIAL_CMD_PLAY_TOGGLE
// note: this partially duplicates CommandLoop; it saves some RAM bytes
// - if wait time elapsed, return true
// - if stopped with button or serial, return false

static bool InterPlayWait(uint32_t pwmCycles)
{
    bool stopped = false;

    //wait for the given amount of time, or until stopped
    while((pwmCycles-- != 0) && (! stopped))
    {
        SyncToPWM();

        //presume no serial command
        uint8_t serialCmd = SERIAL_CMD_NONE;

        if (CheckButtons())
        {
            //Play button pressed: exit delay loop
            stopped = true;
        }
        else
        {
            //no button has been pressed, check for serial command
            serialCmd = GetSerialCommand();

           //execute serial command, if any
           //(will do nothing if a button press has been served)
           switch (serialCmd)
           {
               case SERIAL_CMD_QUERY:
                   //status check: currently playing
                   UART0_SendByte(SERIAL_REPLY_PLAYING);
                   break;

               case SERIAL_CMD_PLAY_TOGGLE:
                   //exit delay loop
                   stopped = true;
                   break;

               case SERIAL_CMD_NEXT:
                   //select next music (index will wrap if out of range)
                   lastBuiltinMusicIdx = SelectBuiltinMusic(lastBuiltinMusicIdx + 1);
                   break;

               case SERIAL_CMD_PREV:
                   //select previous music (index will wrap if out of range)
                   lastBuiltinMusicIdx = SelectBuiltinMusic(lastBuiltinMusicIdx - 1);
                   break;

               default:
                   //all other commands are ignored
                   break;

           } //(switch(serialCmd))
        }
    }

    //return true if time elapsed
    return (! stopped);
}

//-------------------------------------------------------------------------
// execute serial Load command: read music into buffer,
// send back SERIAL_REPLY_OK or error message
// in case of error, restore last built-in music

static void LoadMusicAndReply(void)
{
    //load music from serial
    enum LoadResult r = LoadMusic();
    if (r == LOAD_OK)
    {
        UART0_SendByte(SERIAL_REPLY_OK);
    }
    else
    {
        //load failed: send error message (to be safe, NULL is ignored)
        PGM_P resultString = GetLoadReplyString(r);
        UART0_SendByte(SERIAL_REPLY_ERROR);
        SerialSendPgmString(resultString);
        //music buffer is probably dirty, restore last music
        SelectBuiltinMusic(lastBuiltinMusicIdx);
    }
}

//-------------------------------------------------------------------------
// load music from serial port: dataLenL, dataLenH, data, CRC-8
// (CRC=8 includes dataLenL, dataLenH, data)
// return NULL if OK, else return 0-terminated error message (including '\n')

static enum LoadResult LoadMusic(void)
{
    enum LoadResult result = LOAD_OK;
    uint8_t crc8 = 0;

    union {
        uint16_t val;
        struct { uint8_t L; uint8_t H; }; //assume little endian
    } dataLen;

    //read data length
    dataLen.L = LoadWaitSerialByteIfOk(&crc8, &result);
    dataLen.H = LoadWaitSerialByteIfOk(&crc8, &result);

    //check that there is enough space
    if ((result == LOAD_OK) && (dataLen.val > MAX_MUSIC_LENGTH))
    {
        result = LOAD_ERR_SIZE;
    }

    //read data (ignore if not OK)
    uint8_t *dest;
    for (dest = &musicBuf[0]; (dest < &musicBuf[dataLen.val]) && (result == LOAD_OK); ++dest)
    {
        *dest = LoadWaitSerialByteIfOk(&crc8, &result);
    }

    //read CRC-8 (skip reading if not OK) and check it
    LoadWaitSerialByteIfOk(&crc8, &result);

    if ((result == LOAD_OK) && (crc8 != 0))
    {
        result = LOAD_ERR_CRC;
    }

    return result;
}

//-------------------------------------------------------------------------
// if *result != LOAD_OK do nothing and return 0,
// else wait for a serial byte within LOAD_BYTE_TIMEOUT_MS,
// return byte, update CRC-8, leave *result unchanged if OK,
//        else return 0, do not touch CRC-8, set *result to LOAD_ERR_TIMEOUT

static uint8_t LoadWaitSerialByteIfOk(uint8_t *crc8, enum LoadResult *result)
{
    uint8_t rxByte = 0;

    //just return if there is a previous error in *result
    if (*result == LOAD_OK)
    {
        if (SerialWaitByteWithTimeout(&rxByte, (MS_TO_PWM_CYCLES(LOAD_BYTE_TIMEOUT_MS))))
        {
            *crc8 = _crc_ibutton_update(*crc8, rxByte);
        }
        else
        {
            *result = LOAD_ERR_TIMEOUT;
        }
    }

    return rxByte;
}

//-------------------------------------------------------------------------
// get load reply message, from program memory
// return (pointer to) string, NULL if none
// (could be done in a more compact way, but this is clearer)

static PGM_P GetLoadReplyString(enum LoadResult r)
{
    //pointer to string in program space
    PGM_P ps = NULL;

    if (r == LOAD_ERR_SIZE)
    {
        ps = loadResultErrSize;
    }
    else if (r == LOAD_ERR_CRC)
    {
        ps = loadResultErrCrc;
    }
    else if (r == LOAD_ERR_TIMEOUT)
    {
        ps = loadResultErrTimeout;
    }

    return ps; //NULL if none
};

//-------------------------------------------------------------------------
// check for serial command, convert to lower case
// return received command, or SERIAL_CMD_NONE

static uint8_t GetSerialCommand(void)
{
    uint8_t cmd = SERIAL_CMD_NONE;

    if (UART0_IsRxDataAvailable())
    {
        //a char has been received from the serial port
        cmd = UART0_ReceiveByteNoCheck();
        //convert to uppercase (if applicable)
        cmd = (uint8_t)toupper((char)cmd);
    }

    return cmd;
}

//-------------------------------------------------------------------------
// wait for a byte from UART0 with timeout expressed in PWM cycles (approx)
// return true and byte in *receivedByte if received before timeout,
//        false if timeout, leave *receivedByte unchanged

static bool SerialWaitByteWithTimeout(uint8_t *receivedByte, uint16_t timeoutCycles)
{
    bool ok = false;

    //wait for sync to get the first PWM cycle right
    ResetPWMSync();

    //wait for either data or timeout
    while ((! UART0_IsRxDataAvailable()) && (timeoutCycles > 0))
    {
        SyncToPWM();
        timeoutCycles--;
    }

    if (UART0_IsRxDataAvailable())
    {
        //data received, read it
        *receivedByte = UART0_ReceiveByteNoCheck();
        ok = true;
    }

    return ok;
}

//-------------------------------------------------------------------------
// send a string from program memory to UART0
// do nothing if ps is NULL

static void SerialSendPgmString(PGM_P ps)
{
    if (ps != NULL)
    {
        char c;
        while ((c = pgm_read_byte(ps++)) != '\0')
        {
            UART0_SendByte(c);
        }
    }
}

//-------------------------------------------------------------------------
