//-------------------------------------------------------------------------
// minimalist non-irq handler for UART0
// (non time-critical functions, see userial.h for the macros)
// by Enrico Colombini <erix@erix.it>
// license: CC Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)
// http://creativecommons.org/licenses/by-sa/4.0/
//-------------------------------------------------------------------------

#include "userial.h"
#include <stdint.h>
#include <stdbool.h>
#include <avr/io.h>

//-------------------------------------------------------------------------
// perform a complete (re-)initialization of UART0, 8n1 format
// connect rx,tx pins; leave irqs disabled
// NOTE: assumes no pending tx (could wait for completion but it would block)
//
// the UART0_BAUD_FACTOR macro can be used to precompute baudFactor,
// e.g. for 16 MHz clock, 9600 baud:
//   UART0_Init(UART0_BAUD_FACTOR(16000000, 9600)

void UART0_Init(uint16_t baudFactor)
{
    //disable rx, tx (also free pins) and all irqs, set UCSZ02=0 for 8 bits
    UCSR0B = 0;
    //no 2x, no multi-processor (error bits must be written as 0)
    UCSR0A = 0;
    //asynchronous, 8n1 (UCSZ02 was set above to 0)
    UCSR0C = _BV(UCSZ01) | _BV(UCSZ00);
    //set baud rate
    UBRR0H = (uint8_t)(baudFactor >> 8);
    UBRR0L = (uint8_t)(baudFactor & 0xff);
    //enable UART tx,rx, keep irqs disabled, write 0 into reserved bits
    UCSR0B = (_BV(TXEN0) | _BV(RXEN0));
}

//-------------------------------------------------------------------------
// wait for tx ready, then send data byte
//
//this function is safe but blocks until the transmitter is free
//(check UART0_IsTxBufferFree() before calling to prevent blocking)

void UART0_SendByte(uint8_t byte)
{
    //wait until the transmitter buffer is free
    while (! UART0_IsTxBufferFree())
    {
        //wait
    }
    //send
    UDR0 = byte;
}

//-------------------------------------------------------------------------
// wait for a received byte, return it
//
//this function is safe but blocks until a byte is received, possibly forever
//(check UART0_IsRxDataAvailable() before calling to prevent blocking)

uint8_t UART0_WaitByte(void)
{
    //wait until the receiver has data
    while (! UART0_IsRxDataAvailable())
    {
        //wait
    }
    //read received byte
    return UDR0;
}

//-------------------------------------------------------------------------
